'''
v0.1:
    Now that TraceWin can also convert binary fieldmaps to ASCII the main usage is reduced to trimming fieldmaps
    -- M. Eshraqi 2018 July 09
v0.2:
    Added function to read ASCII fieldmaps and added descriptions
    Added function save_to_ascii
    Added function batch_convert_to_Ascci
    -- M. Eshraqi 2018 July 23

v0.3:
    minor bug fixed in 2D fields
    added function unify_grid_size

v0.4:
    added function create_TW_fieldmap
    added function uniform_mesh
    fixed a bug in function save_to_ascii which was writing Nx, Ny and Nz as floats
    -- M. Eshraqi 2018 Aug 09

---------------------------------------------------------------------------------------------------------

                                            From TraceWin manual:

---------------------------------------------------------------------------------------------------------


        The field map file syntax is the following in the BINARY format:

        - Dimension 1 :
            nz (integer 4 bytes) zmax (double 8 bytes) 
            Norm (double 8 bytes)

                # Nz Zmax
                # Norm

            for k=0 to nz
                Fz(k.zmax/nz) (float 4 bytes)

        - Dimension 2 :
            nz (integer 4 bytes) zmax (double 8 bytes) 
            nx (integer 4 bytes) xmax (double 8 bytes) 
            Norm (double 8 bytes)

                # Nz Zmax
                # Nx Xmax
                # Norm

            for k=0 to nz
                for i=0 to nr
                    Fz(k*zmax/nz, i*rmax/nr) (float 4 bytes)

            - Dimension 3 :
            (Be careful to the dimention order)
            nz (integer 4 bytes) zmax (double 8 bytes)
            nx (integer 4 bytes) xmin (double 8 bytes) xmax (double 8 bytes) 
            ny (integer 4 bytes) ymin (double 8 bytes) ymax (double 8 bytes) 
            Norm (double 8 bytes)
                # Nz Zmax
                # Nx Xmin Xmax
                # Ny Ymin Ymax
                # Norm

            for k=0 to nz
                for j=0 to ny
                    for i=0 to nx
                        Fz(k*zmax/nz, ymin+j*(ymax-ymin)/ny, xmin+i*(xmax-xmin)/nx) (float 4 bytes) 
        Warning: The lattice has to be regular.
        The normalization factor is equal to ke/Norm or kb/Norm.
        Fz are in MV/m for electric field or in T for magnetic field. The dimensions are in meter.

---------------------------------------------------------------------------------------------------------


---------------------------------------------------------------------------------------------------------
'''

__all__ = ['read_binary_fieldmap', 'read_ascii_fieldmap', 'cut_fieldmap_length', 'field_on_axis_from_file'
           'save_to_ascii', 'batch_convert_to_ascii', 'unify_grid_size', 'batch_cut_fieldmap',
           'uniform_mesh', 'create_TW_fieldmap']

#class FieldMapy:
#'''
#class contains functions needed for manipulating fieldmaps for TraceWin.
#'''
#__version__ = "0.4"

import os
import numpy as np

def read_binary_fieldmap(path_plus_file: str, fieldmap_dim: int):
    '''
    parameters
    ----------
    path_plus_file: string
        path to the binary fieldmap file including the path, file name and extension
    fieldmap_dim: int
        dimension of the fieldmap, 1, 2 and 3 for 1D, 2D and 3D TraceWin formatted fieldmap

    returns
    -------
    a list containing Header and Field
    Header: 1D array of size 3, 5, 8 (for 1D, 2D and 3D fieldmaps)
        TraceWin fieldmap header info
    Field: 1D numpy array containing the nD field data
    '''
    field_File = open(path_plus_file,'r')

    if fieldmap_dim == 3:
        Header_Type = np.dtype([('Nz', 'int32'), ('Zmax', 'float64'), ('Nx','int32'), ('Xmin', 'float64'), ('Xmax', 'float64'), 
                     ('Ny','int32'), ('Ymin','float64'), ('Ymax','float64'), ('Norm', 'float64')])
        THeader = np.fromfile(field_File, dtype=Header_Type, count=1)
        Nz = THeader['Nz'][0]
        Zmax = THeader['Zmax'][0]
        Nx = THeader['Nx'][0]
        Xmin = THeader['Xmin'][0]
        Xmax = THeader['Xmax'][0]
        Ny = THeader['Ny'][0]
        Ymin = THeader['Ymin'][0]
        Ymax = THeader['Ymax'][0]
        Norm = THeader['Norm'][0]
        Header = np.zeros(len(THeader))
        Header = np.array([Nz, Zmax, Nx, Xmin, Xmax, Ny, Ymin, Ymax, Norm])

    elif fieldmap_dim == 2:
        Header_Type = np.dtype([('Nz', 'int32'), ('Zmax', 'float64'), ('Nx','int32'), ('Xmax', 'float64'), 
                     ('Norm', 'float64')])
        THeader = np.fromfile(field_File, dtype=Header_Type, count=1)
        Nz = THeader['Nz'][0]
        Zmax = THeader['Zmax'][0]
        Nx = THeader['Nx'][0]
        Xmax = THeader['Xmax'][0]
        Norm = THeader['Norm'][0]
        Header = np.zeros(len(THeader))
        Header = np.array([Nz, Zmax, Nx, Xmax, Norm])

    elif fieldmap_dim == 1:
        Header_Type = np.dtype([('Nz', 'int32'), ('Zmax', 'float64'), ('Norm', 'float32')])
        Header = np.fromfile(field_File, dtype=Header_Type, count=1)
        Nz = THeader['Nz'][0]
        Zmax = THeader['Zmax'][0]
        Norm = THeader['Norm'][0]
        Header = np.zeros(len(THeader))

    Field_Type = np.dtype([('F_val','float32')])  

    Nf = int(Nz+1)
    if fieldmap_dim == 2:
        Nf *= int(Nx+1)
    elif fieldmap_dim == 3:
        Nf *= (int(Ny+1)*int(Nx+1))

    Field = np.fromfile(field_File, dtype=Field_Type, count=Nf)

    return Header, Field

def read_ascii_fieldmap(path_plus_file, fieldmap_dim: int):
    '''
    parameters
    ----------
    path_plus_file: string
        path to the ASCII fieldmap file including the path, file name and extension
    fieldmap_dim: int
        dimension of the fieldmap, 1, 2 and 3 for 1D, 2D and 3D TraceWin formatted fieldmap

    returns
    -------
    a list containing Header and Field
    Header: 1D array of size 3, 5, 8 (for 1D, 2D and 3D fieldmaps)
        TraceWin fieldmap header info
    Field: 1D numpy array containing the nD field data
    '''

    from itertools import islice
    import numpy as np

    items_per_row = np.array([[2, 1, 0, 0], 
                              [2, 2, 1, 0],
                              [2, 3, 3, 1]]) # for 1D, 2D and 3D fields in an ASCII fieldmap in TraceWin format
    l_header = np.sum(items_per_row[fieldmap_dim-1,:], axis=0)  # getting the length of the header from the items
    Header = np.zeros(l_header)

    j = 0
    with open(path_plus_file, 'r') as f:
        for i in range(0, fieldmap_dim+1):
            header_line = list(islice(f, 1))
            fl = header_line[0]
            fl = fl.rstrip('\n')
            fl = fl.split(' ')
            l_header_line = items_per_row[int(fieldmap_dim)-1, i]
            Header[j:j+l_header_line] = fl[:]
            j+=l_header_line

        Nf = int(Header[0]+1)
        if fieldmap_dim == 2:
            Nf *= int(Header[2]+1)
        elif fieldmap_dim == 3:
            Nf *= int(Header[2]+1)*int(Header[5]+1)

        fieldmap = np.zeros(int(Nf))

        i = 0
        for line in f:
            eorb = line.rstrip('\n')
            fieldmap[i] = eorb[:]
            i+=1

    return Header, fieldmap

def cut_fieldmap_length(path_plus_file: str, fieldmap_dim: int, entrance_rows_cut: int, 
                        exit_rows_cut: int, file_type: str):
    '''
    parameters
    ----------
    path_plus_file: string
        path to the fieldmap file (ASCII or binary) including the path, file name and extension
    fieldmap_dim: int
        dimension of the fieldmap, 1, 2 and 3 for 1D, 2D and 3D TraceWin formatted fieldmap
    entrance_rows_cut: int
        number of data rows to be removed from array from the entrance side of the field
    exit_rows_cut: int    
        number of data rows to be removed from array from the exit side of the field
    file_type: str
        'b' for binary and 'a' for ascii fieldmap files.

    returns
    -------
    a list containing Header and Field
    Header: 1D array of size 3, 5, 8 (for 1D, 2D and 3D fieldmaps)
        TraceWin fieldmap header info
    Field: 1D numpy array containing the nD field data with only the remaining rows


    raises
    ------
    prints a message if the file_type is not specified to be a (ascii) or b (binary)

    missing
    -------
    cutting the field in other directions    
    '''

    if file_type == 'a': 
        Field, Header = read_ascii_fieldmap(path_plus_file, fieldmap_dim)
    elif file_type == 'b':
        Field, Header = read_binary_fieldmap(path_plus_file, fieldmap_dim)
    else:
        print('file_type should be a for ascii or b for binary')

    lowx = np.arange(0,int(entrance_rows_cut))  #list of rows to be removed from the entrance side
    highx = np.arange(int(Header[0]+1-exit_rows_cut), int(Header[0]+1)) # list of rows to be removed from the exit side

    newNz = (Header[0]+1)-entrance_rows_cut-exit_rows_cut # new number of field data points after removal of rows
    stepZ = Header[1] / Header[0]

    # depending the dimension of the field the fieldmap is reshaped for line removal
    if fieldmap_dim == 1:
        newlength = newNz
    elif fieldmap_dim == 2:
        Fieldmapmatrix = np.reshape(Field, (int(Header[0]+1), int(Header[2]+1)))
        newlength = newNz*(Header[2]+1)
    elif fieldmap_dim == 3:
        Fieldmapmatrix = np.reshape(Field, (int(Header[0]+1), int((Header[2]+1)*(Header[5]+1))))
        #print(Fieldmapmatrix.shape)
        newlength = newNz*(Header[2]+1)*(Header[5]+1)

    if (entrance_rows_cut>0 and exit_rows_cut>0):
        x = np.delete(Fieldmapmatrix,(lowx, highx), axis=0)
    elif (entrance_rows_cut ==0):
        x = np.delete(Fieldmapmatrix,(highx), axis=0)
    elif exit_rows_cut ==0:
        x = np.delete(Fieldmapmatrix,(lowx), axis=0)

    # reshaping the array back to a 1D array
    ShortenedMatrix = np.reshape(x, (int(newlength), 1))

    # creating new header data (if other directions are added to the cut fuction the header for those have 
    # to be rewritten too)
    Header[0] = newNz-1
    Header[1] = stepZ*(newNz-1)

    return Header, ShortenedMatrix

def field_on_axis_from_file(path_plus_file, fieldmap_dim: int, file_type: str):
    '''
    parameters
    ----------
    path_plus_file: string
        name of the fieldmap file including its path and extension 
    fieldmap_dim: int
        dimesion of the fieldmap, 1, 2, 3 for 1D, 2D and 3D fieldmaps in TraceWin format
    file_type: string
        'b' for binary and 'a' for ascii fieldmap files.

    returns
    -------
    returns a 2D array containing the z coordinates and the field on axis in the z direction

    Take a 2D or 3D fieldmap as a 1D numpy array (as from TraceWin) in MV/m, 
    dimension of the fieldmap, 
    returns the field on-axis in the z direction
    example:
    field_on_axis_from_file('Data/FM/HB_W_coupler.edz', 3, 'b')
    '''
    from numpy import zeros as npzeros

    if file_type == 'b':
        Header, Field = read_binary_fieldmap(path_plus_file, fieldmap_dim)
    elif file_type == 'a':
        Header, Field = read_ascii_fieldmap(path_plus_file, fieldmap_dim)
    else:
        print('file_type should be a for ascii or b for binary')

    if fieldmap_dim == 2:
        Nz, Nrx = Header[0], Header[2]
        Fieldmapmatrix = np.reshape(Field, (int(Nz+1), int(Nrx+1)))
        midpoint = 0
    elif fieldmap_dim == 3:
        Nz, Nrx, Ny = Header[0], Header[2], Header[5]
        Fieldmapmatrix = np.reshape(Field, (int(Nz+1), int((Nrx+1)*(Ny+1))))
        midpoint = int(Nrx*Ny/2)

    dz = Header[1]/Nz
    field_on_axis = npzeros((2,int(Nz)+1))
    for i in range(int(Nz+1)):
        field_on_axis[0,i] = i*dz

    field_on_axis[1,:] = Fieldmapmatrix[:, midpoint]
    return field_on_axis

def save_to_ascii(Header, Field, fieldmap_dim: int, path_to_out_file: str, out_file_name: str):
    '''
    Parameters
    ----------
    Header: array like
        1D array of size (3, 5 or 8) for 1D, 2D or 3D fieldmap header info, TraceWin format
    Field: array like
        1D array of size Nz fieldmap containing field data
    fieldmap_dim: int
        dimesion of the fieldmap, 1, 2, 3 for 1D, 2D and 3D fieldmaps in TraceWin format
    path_to_out_file: string
        path to the output file folder
    out_file_name: string
        name of the outputfile, including the extension

    Returns
    -------
    None

    What does it do
    ---------------
    Saves a TraceWin readable ASCII file
    '''
    with open(os.path.join(path_to_out_file, out_file_name), 'w') as out_file:

        stepZ = Header[1] / Header[0]   #stepZ = Zmax / Nz

        out_file.write(str(int(Header[0])))
        out_file.write(' ')
        out_file.write(str(Header[1]))
        #out_file.write(' '.join(str(hval) for hval in Header[0:2]))
        out_file.write('\n')
        if fieldmap_dim == 1:
            #Norm = Header[2]
            out_file.write((str(Header[2])))
            out_file.write('\n')

        elif fieldmap_dim == 2:
            #Norm = Header[4]
            stepX = Header[3]/Header[2]  #stepX = Xmax / Nx

            out_file.write(str(int(Header[2])))
            out_file.write(' ')
            out_file.write(str(int(Header[3])))
            #out_file.write(' '.join(str(hval) for hval in Header[2:4]))
            out_file.write('\n')
            out_file.write(str(Header[4]))
            out_file.write('\n')

        elif fieldmap_dim == 3:
            #Norm = Header[8]
            stepX = (Header[4]-Header[3])/Header[2]  #stepX = (Xmax - Xmin) / Nx
            stepY = (Header[7]-Header[6])/Header[5]  #stepY = (Ymax - Ymin) / Ny

            out_file.write(str(int(Header[2])))
            out_file.write(' ')
            out_file.write((' '.join(str(hval) for hval in Header[3:5])))
            out_file.write('\n')
            out_file.write(str(int(Header[5])))
            out_file.write(' ')
            out_file.write((' '.join(str(hval) for hval in Header[6:8])))
            out_file.write('\n')
            out_file.write(''.join(str(Header[8])))
            out_file.write('\n')

        for fval in Field:
            try:
                out_file.write(str(fval[0]))
                out_file.write('\n')#print(fval)
            except IndexError:
                out_file.write(str(fval))
                out_file.write('\n')#print(fval)

    return

def batch_convert_to_ascii(binary_field_path: str, fieldmap_dim: int):
    '''
    parameters
    ----------
    binary_field_path: string
                  path to the binary files to be converted to ASCII format
    fieldmap_dim: int
                  dimension of the fieldmap to be converted
    '''
    onlyfiles = [f for f in os.listdir(binary_field_path) if os.path.isfile(os.path.join(binary_field_path, f))]

    for f in onlyfiles:
        if f[0]!=".":

            Header, Field = read_binary_fieldmap(binary_field_path+f, fieldmap_dim)
            save_to_ascii(Header, Field, fieldmap_dim, binary_field_path+'/ASCII/', f)    

    return

def unify_grid_size(old_x: list, old_y: list, new_x: list, **kwargs):
    '''
    parameters:
    old_x: list, 
    old_y: list, 
    new_x: list

    **kwargs: kind
    ‘linear’, 
    ‘nearest’, 
    ‘zero’, ‘slinear’, ‘quadratic’, ‘cubic’: refer to a spline interpolation of zeroth, first, second or third order
    ‘previous’, ‘next’: return the previous or next value of the point
    An integer specifying the order of the spline interpolator to use. Default is ‘linear’.

    Retunrs:
    -------
    new_y: list

    Takes a mesh as x values and the corresponding values of a function.
    creates a spline interpolation for those data sets.
    Using a new_x it returns the correcposnding new_y.
    Could be used to convert an irregular mesh to a regular mesh or 
    to increase the number of mesh points
    '''
    kind = kwargs.get('kind', 'linear')
    save_to_file = kwargs.get('save_to_file', 'none')

    from scipy import interpolate
    f = interpolate.interp1d(old_x, old_y, kind)
    new_y = f(new_x)
    if save_to_file != 'none':
        field = np.zeros((2, len(new_x)))
        field[0,:] = new_x
        field[1,:] = new_y
        np.savetxt(save_to_file, np.transpose(field), delimiter='\t')

    return  new_y

def batch_cut_fieldmap(fieldmaps_path: str, fieldmap_dim: int, entrance_rows_cut: int, 
                        exit_rows_cut: int, file_type: str):
    '''
    parameters
    ----------
    fieldmaps_path: string
        path to the fieldmap folder (ASCII or binary) excluding the trailing / and filename
    fieldmap_dim: int
        dimension of the fieldmap, 1, 2 and 3 for 1D, 2D and 3D TraceWin formatted fieldmap
    entrance_rows_cut: int
        number of data rows to be removed from array from the entrance side of the field
    exit_rows_cut: int    
        number of data rows to be removed from array from the exit side of the field
    file_type: str
        'b' for binary and 'a' for ascii fieldmap files.

    Returns
    -------
    None

    What does it do
    ---------------
    Saves a TraceWin readable ASCII file
    '''
    onlyfiles = [f for f in os.listdir(fieldmaps_path) if os.path.isfile(os.path.join(fieldmaps_path, f))]

    for f in onlyfiles:
        if f[0]!=".":
            if (entrance_rows_cut+exit_rows_cut>0):
                Header, Field = cut_fieldmap_length(f, fieldmap_dim, entrance_rows_cut, exit_rows_cut, file_type)

        save_to_ascii(Header, Field, fieldmap_dim, fieldmaps_path+'/ASCII/', f)    

def uniform_mesh(mesh: list):
    '''
    parameters: list
        a 1D array which is the coordinate values of the field
    returns: Tuple
        a tuple containing a boolean (if the mesh is uniform or not), mesh_size (average to reduce the effect of 
        floating point error), and the number of steps from min to max (n numbers would result in n-1)
    '''
    unique = np.unique(mesh, return_counts=False)
    mesh_check = np.zeros(len(unique)-1)
    for i in range(1,len(unique)):
        mesh_check[i-1]=(unique[i]-unique[i-1])
    mesh_check, mesh_size = np.unique(mesh_check, return_counts=True)

    if np.abs(max(mesh_check))>np.abs(min(mesh_check))*1.001:
        print('Mesh size differs by more than 0.1%, should be fixed')
        return False, mesh_check, sum(mesh_size)
    else:
        return True, np.average(mesh_check), sum(mesh_size)

def create_TW_fieldmap(fieldmaps_path: str, tabular_file_name: str, fieldmap_dim: int = 1, n_col: int = 1, header: bool = True, **kwargs):
    '''
    paramters
    ---------
    fieldmaps_path: str
        path to the fieldmap 
    tabular_file_with_path: file name
        tabular data with coordinates z, x and y or z and r or only z depending on dimension as first 1-3 cols.
        n_col is the number of data columns except coordinate ones (1-6)
        coordinates should be on a uniform grid on each axis (but different axes are independent)
    fieldmap_dim: int
        dimension of the fieldmap data, default is 1
    n_col:  int
        number of data columns in fieldmap, default is 1
    header: boolean
        if the file has header or not, default is True
    **kwargs
        extensions: list
            list of column names to be used for naming the files, default is [F1, F2, F3, F4, F5, F6]
        xyz: list
            order of columns, xyz = [col_number(x), col), col_number(y), col_number(z)]
                e.g.if column 1 is x, col 2 is y and col 3 is z, xyz list = [1,2,3] (default)
        delimeter: string
            delimeter used in the data, default is tab '\t'
    returns
    -------
    fieldmaps in the TraceWin format, one per n_col.

    '''
    extensions = kwargs.get('extensions', ['F1', 'F2', 'F3', 'F4', 'F5', 'F6'])
    xyz = kwargs.get('xyz', [1, 2, 3])
    delim = kwargs.get('delimeter', '\t')

    if len(xyz)<fieldmap_dim:
        print('xyz array is shorter than the number of dimensions')
        return

    tabular_file_with_path = fieldmaps_path + tabular_file_name
    field_File = open(tabular_file_with_path,'r')

    if header==True:
        field_array = np.genfromtxt(field_File, delimiter=delim, skip_header=1)
    else:
        field_array = np.genfromtxt(field_File, delimiter=delim, skip_header=0)
    field_File.close()

    items_per_row = np.array([[2, 1, 0, 0], 
                              [2, 2, 1, 0],
                              [2, 3, 3, 1]]) # for 1D, 2D and 3D fields in an ASCII fieldmap in TraceWin format
    l_header = np.sum(items_per_row[fieldmap_dim-1,:], axis=0)  # getting the length of the header from the items
    Header = np.zeros(l_header)

    # sort first on x coordinate, then y and then z for TraceWin fieldmap
    for i in range(0, fieldmap_dim):
        coor = field_array[:,xyz[i]-1]
        chk_mesh, d_mesh, n_mesh = uniform_mesh(coor)
        if chk_mesh:
            if i == fieldmap_dim - 1:
                Header[0] = n_mesh
                Header[1] = max(coor)
                # \n
            if fieldmap_dim == 1:
                Header[2] = 1
                # \n
            elif fieldmap_dim == 2:
                if i == fieldmap_dim - 2:
                    Header[2] = n_mesh
                    Header[3] = max(coor)
                    # \n
                    Header[4] = 1
                    # \n
            elif fieldmap_dim == 3:
                if i == fieldmap_dim - 2:
                    Header[2] = n_mesh
                    Header[3] = min(coor)
                    Header[4] = max(coor)
                    # \n
                elif i == fieldmap_dim - 3:
                    Header[5] = n_mesh
                    Header[6] = min(coor)
                    Header[7] = max(coor)
                    # \n
                    Header[8] = 1
                    # \n

            else: # fieldmap_dim > 3 or fieldmap_dim < 1:
                print('Invalid dimension, should be 0, 1 or 2')
                return
        ordr = np.argsort(coor, kind = 'mergesort')
        field_array = field_array[ordr, :]

    #print('\n', field_array[:,:3])
    #print(Header)
    filename_wo_ext = os.path.splitext(tabular_file_name)[0]
    for i in range(fieldmap_dim, fieldmap_dim+n_col):
        #print(len(field_array[:,i]))
        save_to_ascii(Header, field_array[:,i], fieldmap_dim, fieldmaps_path, filename_wo_ext+"."+extensions[i-fieldmap_dim])    

#  - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - #
#                                                                              #
#            C h a n g e   t h e   f o l l o w i n g    v a l u e s            #
#                                                                              #
#  - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - #

def example_func():
    '''
    Example

    The parameters in the following six rows define the path to your fieldmap, no name is needed, this script 
    scans the folder and converts all files, assumes all are binary fieldmaps, to ASCII fieldmaps.
    The dimension of the field is also needed.
    If you need to cut the field, the n_enter and n_exit should be adjusted.
    '''
    mypath = 'Data/FM/'          # where the fieldmaps are stored
    fieldmap_dim = 3
    #fieldmap_dim = 2
    #fieldmap_dim = 1
    n_enter = 0                 #Number of lines to be removed from entrance of fieldmap
    n_exit = 0                  #Number of lines to be removed from exit of fieldmap

    onlyfiles = [f for f in os.listdir(mypath) if os.path.isfile(os.path.join(mypath, f))]

    for f in onlyfiles:
        if f[0]!=".":

            if (n_enter+n_exit>0):
                Header, Field = cut_fieldmap_length(f, fieldmap_dim, n_enter, n_exit, 'b')
            else:
                Header, Field = read_binary_fieldmap(mypath+f, fieldmap_dim)

            save_to_ascii(Header, Field, fieldmap_dim, 'Data/FM/ASCII/', f)

    test = False
    #%pylab inline
    if test == True:
        #field = field_on_axis_from_file('Data/FM/MB_W_coupler.edz', 3, 'b')
        field = field_on_axis_from_file('Data/FM/ASCII/Spoke_W_coupler.edz', 3, 'a')
        plot(field[0,:], field[1,:])
